function [ emd, flow, lpsolver_out ] = emd(sig1, sig2, D)
%EMD Computes EMD [Rubner et al, 2000] between two signatures
%
% Input:
%
%   sig1 - first signature - a vector of weights; |sig1| = n
%   sig2 - second signature; |sig2| = m
%   D - ground distance; an n-by-m matrix, D(i,j) - cost of moving
%       a unit of weight from i'th item in sig1 to j'th item in sig2
%
% Output:
%
%   dist - a real value representing the EMD between sig1 and sig2
%   flow - optimal flow of mass between the signatures
%
%#ok<*NASGU>

%% Input validation

if(~isvector(sig1) || ~isvector(sig2) || ~ismatrix(D))
    error('Invalid input.');
end

[n, m] = size(D);
if(size(sig1, 1) ~= n)
    sig1 = sig1';
end
if(size(sig2, 1) ~= m)
    sig2 = sig2';
end
if(size(sig1, 1) ~= n || size(sig2, 1) ~= m)
    error('Signature sizes do not agree with the size of D.');
end

%% Definition of objective function and constraints

% Objective function's coefficients
Dex = reshape(D, n * m, 1);

% Inequality-type constraints (supply/demand limits):
A = sparse([ kron(ones(1, m), speye(n, n)); kron(eye(m, m), ones(1, n)) ]);
b = sparse([ sig1; sig2 ]);

% Equality-type constraints (desirable total flow):
Aeq = ones(1, n * m);
min_mass = min(sum(sig1), sum(sig2));
beq = min_mass;

% Lower bounds (non-negative flow):
lb = zeros(n * m, 1);

%% Computing EMD

% Choosing solver
if(cplex_available())
    solver_name = 'cplexlp';
else
    solver_name = 'linprog';
end

[consout, flow, emd, exitcode, output] = evalc([solver_name, '(Dex, A, b, Aeq, beq, lb)']);

% Logging
lpsolver_out = struct('consout', consout, 'exitcode', exitcode, 'output', output);

% Preparing output
flow = reshape(flow, n, m);
emd = emd / min_mass;

end