function [ dist, flows, log ] = snd(P, Q, adjmat, cost_model, gamma, options)
%%SND2 Computes a Social Network Distance (SND) between two polar opinion
%	   distributions (network states) P and Q over a sparse network.
%
%	   This is a non-symmetric version of SND; a symmetric version of SND (as
%	   it was defined in the paper [1]) is 0.5 * (snd(P, Q, ...) + snd(Q, P, ...)).
%
% INPUT:
%
%   P -- first polar opinion distribution (|P| = n, P(i) in {-1, 0, +1})
%
%   Q -- second polar opinion distribution
%
%   adjmat -- sparse n-by-n adjacency matrix of the underlying network. By
%       default, it is assumed to be the transpose (columns are sources, rows
%       are destinations). If you use a custom shortest path solver, specified in
%       options.sp_solver, supply adjmat in whether form your solver accepts it.
%
%   cost_model -- a structure containing information for generating opinion
%       propagation costs, with the following fields:
%
%       params -- parameters of the model
%
%       func -- cost generating function
%
%           [ Aext ] = @(model_params, adjmat, user_states, op, transpose)
%
%               Aext -- opinion propagation costs matrix that incorporates topological
%                   costs ("communication costs"), costs for spreading, and costs for adoption
%               model_params -- parameters of the model
%               adjmat -- adjacency matrix of the network
%               user_states -- vector of user opinions
%               op -- opinion being transported
%               transpose -- whether compute original or transposed cost matrices
%
%       Default cost_model is snd_simple_cost_model(NaN).
%
%   gamma -- ground distance to/from the banks. If the adjmat is extended
%       with the banks, then gamma are the costs on the newly create edges
%       to/from the bank nodes (bins). Default is 1.
%
%   options -- implementation options
%
%       options('return_flow') -- logical indicating whether to return the
%           optimal flow(s) in addition to the distance value. If true, the
%           second output argument should be provided. Default is false (recommended).
%
%       options('mcf_solver') -- function [cost, flow, log](supplies, demands, D)
%           serving as a minimum-cost flow solver. Output arguments flow and log
%           are optional. Currently, it is either @bp_mincost_maxflow based on the
%           general-purpose LP-solver from either CPLEX (preferred, if available)
%			or MATLAB, or @gt_transport based on implementation CS2 of Goldberg-Tarjan's
%			min-cost flow algorithm (default).
%
%       options('mcf_precision') -- optional structure describing precision
%           (# of frac. digits) for computing min-cost flow:
%               mcf_precision.cap -- precision for node capacities
%               mcf_precision.cost -- precision of transportation costs
%           If not specified, the default MCF-solver @gt_transport chooses
%           the precision at its discretion.
%
%       options('sp_solver') -- function [dists](adjmat, src) used to find the
%           lengths (dists) of all shortest paths from src to every other node
%           in the graph. Default is Dijkstra with radix heap (@radix_dijkstra)
%           that assumes that the supplied adjmat is transposed.
%
%       options('reduce_dists') -- logical value indicating whether to remove
%           some bins from the distributions without affecting the solution, but
%           reducing the problem's size and, hence, the computation time.
%           Default is true.
%
%       options('debug_mode') -- logical value indicating (verbose) debug mode
%
% OUTPUT:
%
%   dist -- the computed distance value
%
%   flows -- a cell array of flow matrices, one for each internal computation of EMD.
%		Besides specifying this output parameter, one needs to provide the option
%		options('return_flow') equal True (otherwise, flows will not be filled). Note,
%		that each flow matrix also includes flows from artificial nodes ("the banks").
%
%   log -- optional information useful for debugging / profiling
%
% EXAMPLES:
%
%	See snd_example.m, snd_example_discon_net.m.
%
% AUTHOR:
%
%	Anonymized
%
% REFERENCES:
%
%	Anonymized
%

%% Validating number of arguments
if(nargin < 4)
    error('Not enough arguments provided.');
end
%% Validating distributions
if(~issparse(P) || ~issparse(Q))
    error('Either P or Q (or both) are dense. Sparse distributions are expected.');
end
n = size(P, 1);
if(size(P, 2) > 1)
    if(n > 1)
        error('P has invalid size.');
    end
    P = P';
    n = length(P);
end
if(size(Q, 1) ~= n)
    if(size(Q, 2) ~= n)
        error('Q has invalid size.');
    end
    Q = Q';
elseif(size(Q, 2) > 1)
    error('Q has invalid size.');
end
%% Validating adjacency matrix
if(~issparse(adjmat))
    error('Adjacency matrix is supposed to be sparse.');
end
if(size(adjmat, 1) ~= size(adjmat, 2))
    error('Adjmat is not square.');
elseif(size(adjmat, 1) ~= n)
    error('Adjmat''s size does not match the sizes of distributions P and Q.');
end
%% Validating cost model
if(~isstruct(cost_model))
    cost_model = snd_simple_cost_model(NaN);
elseif(~isfield(cost_model, 'func') || ~isa(cost_model.func, 'function_handle'))
	error('The cost model lacks the cost generating function.');
end
%% Validating gamma
if(nargin < 5 || isnan(gamma))
    gamma = 1;
elseif(gamma < 0)
    error('Invalid bank ground distance %f.', gamma);
end
%% Validating options
if(~exist('options', 'var') || ~isstruct(options))
    options = struct();
elseif(~isstruct(options))
    error('Options is supposed to be a structure.');
end
if(~isfield(options, 'return_flow'))
    options.return_flow = false;
elseif(options.return_flow && nargout < 2)
    error('options.return_flow is true, yet, the second output argument is not supplied.');
end
if(~isfield(options, 'mcf_solver'))
    options.mcf_solver = @gt_transport;
end
if(~isfield(options, 'mcf_precision'))
    options.mcf_precision = NaN;
end
if(~isfield(options, 'sp_solver'))
    options.sp_solver = @radix_dijkstra;
end
if(~isfield(options, 'reduce_dists'))
    options.reduce_dists = true;
end
if(~isfield(options, 'debug_mode'))
    options.debug_mode = false;
end
if(length(fieldnames(options)) > 6)
    error('Some options have not been recognized.');
end
%% Output arguments
logging = nargout >= 3;

%% Computing SND

dist = 0;
if(options.return_flow || nargout >= 2); flows = { }; end
if(logging); log.timing = struct('total', NaN, 'shpaths', [ ], 'mcf', [ ]); end

% Splitting the flow by sentiment type
for sent_type = [ +1, -1 ]

    %% P_st -- P with only sentiment of type sent_type present
    P_st = P;
    P_st(sign(P_st) ~= sign(sent_type)) = 0;
    P_st = abs(P_st);

    %% Q_st -- Q with only sentiment of type sent_type present
    Q_st = Q;
    Q_st(sign(Q_st) ~= sign(sent_type)) = 0;
    Q_st = abs(Q_st);

    %% Computing mass mismatch
    P_st_mass = sum(P_st);
    Q_st_mass = sum(Q_st);
    mass_mismatch = abs(P_st_mass - Q_st_mass);

    %% SND to/from an empty histogram from/to an non-empty histogram
    if(xor(P_st_mass == 0, Q_st_mass == 0))
        dist = Inf;
        log = 'SND to an empty histogram is undefined (+Inf).';
        return;
    end

    %% Defining bank capacities (only the lighter histogram gets non-empty banks)
    if(mass_mismatch > 0 && P_st_mass < Q_st_mass)
        P_banks = mass_mismatch * P_st ./ P_st_mass;
    else
        P_banks = [ ];
    end
    if(mass_mismatch > 0 && Q_st_mass < P_st_mass)
        Q_banks = mass_mismatch * Q_st ./ Q_st_mass;
    else
        Q_banks = [ ];
    end

    %% Choosing sources and destinatoins for mass transportation:
    %   (i) choosing only non-empty bins
    %  (ii) choosing only bins whose values are different in P and Q
    if(options.reduce_dists)
        bin_value_mismatch_indic = (P_st ~= Q_st);
        if(nnz(bin_value_mismatch_indic) == 0)
            continue;
        end
    else
        bin_value_mismatch_indic = true;
    end

    % P_st's sources are: (i) all non-empty bins if banks are present, or
    % (ii) all non-empty bins whose value in Q_st is different if there are
    % no banks.
    if(~isempty(P_banks))
        P_st_bin_indic = P_st > 0;
    else
        P_st_bin_indic = (P_st > 0) & bin_value_mismatch_indic;
    end
    % Similarly, for Q_st.
    if(~isempty(Q_banks))
        Q_st_bin_indic = Q_st > 0;
    else
        Q_st_bin_indic = (Q_st > 0) & bin_value_mismatch_indic;
    end

    %% Obtaining the cost matrix
    adjmat_w_spc = cost_model.func(cost_model.params, adjmat, P, sent_type, nnz(P_st_bin_indic) > nnz(Q_st_bin_indic));

    %% Computing ground distance (new version)

    % Compute shortest paths from (some of the) originally present bins of P_st
    % to (some of the) originally present bins of Q_st
    if(logging); t0 = tic; end
    D1 = mssp(find(P_st_bin_indic), find(Q_st_bin_indic), adjmat_w_spc, options.sp_solver);
    if(max(max(D1)) >= +Inf)
        error('Some nodes are unreachable (some D(i,j) = +Inf).');
    end
    % Add shortest paths either from or to the bank bins.
    if(~isempty(P_banks))
        D = D1(get_masked_idx(P_st_bin_indic, bin_value_mismatch_indic), :);
    elseif(~isempty(Q_banks))
        D = D1(:, get_masked_idx(Q_st_bin_indic, bin_value_mismatch_indic));
    else
        D = D1;
    end
    if(~isempty(P_banks))
        D = [ D; D1 + gamma ]; %#ok<AGROW>
    elseif(~isempty(Q_banks))
        D = [ D, D1 + gamma ]; %#ok<AGROW>
    end
    clear D1;
    if(logging); log.timing.shpaths(length(log.timing.shpaths) + 1) = toc(t0); end

    %% Computing optimal flow
    if(logging); t1 = tic; end;
    mcf_from = full([ P_st((P_st > 0) & bin_value_mismatch_indic); P_banks(P_banks > 0) ]);
    mcf_to = full([ Q_st((Q_st > 0) & bin_value_mismatch_indic); Q_banks(Q_banks > 0) ]);
    if(options.return_flow)
        if(options.debug_mode)
            [cost, flow] = options.mcf_solver(mcf_from, mcf_to, D, NaN, NaN, false, true);
        elseif(isequal(options.mcf_solver, @gt_transport))
            [cost, flow] = options.mcf_solver(mcf_from, mcf_to, D, -1, options.mcf_precision, true, false);
        else
            [cost, flow] = options.mcf_solver(mcf_from, mcf_to, D);
        end
        P_idx_range1 = 1 : length(P_st);
        P_idx_range1 = P_idx_range1((P_st > 0) & bin_value_mismatch_indic);
        P_idx_range = [ P_idx_range1, (1 : nnz(P_banks)) + length(P_st) ];
        clear P_idx_range1;
        Q_idx_range1 = 1 : length(Q_st);
        Q_idx_range1 = Q_idx_range1((Q_st > 0) & bin_value_mismatch_indic);
        Q_idx_range = [ Q_idx_range1, (1 : nnz(Q_banks)) + length(Q_st) ];
        clear Q_idx_range1;
        flow = spmat_thinout(flow, length(P_st) + nnz(P_banks), length(Q_st) + nnz(Q_banks), P_idx_range, Q_idx_range);
        flows{length(flows) + 1} = struct('sent_type', sent_type, 'flow', flow);
    else
        if(options.debug_mode)
            cost = options.mcf_solver(mcf_from, mcf_to, D, NaN, NaN, false, true);
        elseif(isequal(options.mcf_solver, @gt_transport))
            cost = options.mcf_solver(mcf_from, mcf_to, D, -1, options.mcf_precision, true, false);
        else
            cost = options.mcf_solver(mcf_from, mcf_to, D);
        end
    end
    if(logging); log.timing.mcf(length(log.timing.mcf) + 1) = toc(t1); end;

    dist = dist + cost;
end

if(logging); log.timing.total = sum(log.timing.shpaths) + sum(log.timing.mcf); end

end

%% Sparse matrix thin-out
function [ A_new ] = spmat_thinout(A, m, n, row_idx, col_idx)
    A_new = sparse([], [], [], m, n, nnz(A));
    A_new(row_idx, col_idx) = A;
end

%% Row/column index masking
function [ idx ] = get_masked_idx(col_indic, col_mask)

v = sparse(size(col_indic, 1), size(col_indic, 2));
v(col_indic) = 1 : nnz(col_indic);
v = v .* col_mask;
idx = v(v > 0);

end

