#include <cstdlib>
#include <cstring>
#include <stdio.h>
#include <stdint.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <cmath>
#include <vector>
#include "mex.h"

using std::string;
using std::stringstream;
using std::ostringstream;
using std::ofstream;
using std::vector;

string get_mex_dirpath();
bool parse_cs2_output(const string &cs2_raw_output, mxArray** plhs0, mxArray** plhs1, uint64_t nlhs);
string run_cs2(const string &input_fname, const double scale_factor, bool need_flow, bool debug_mode);
void trans_to_dimacs(ofstream &os, const mxArray *sup, const mxArray *dem, const mxArray *D);
string write_str_to_tmpfile(string text);

template<typename T> inline string to_stringx(T x);
inline double stodx(string str);
inline uint64_t d2uint64(double x);

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
	if(nrhs < 3) {
		mexErrMsgTxt("Not enough arguments.");
		return;
	}
    if(nlhs < 1 || nlhs > 2) {
        mexErrMsgTxt("This implementation must have 1 or 2 output argument.");
        return;
    }

	for(uint64_t i = 0; i < nrhs; ++i) {
		if(!mxIsDouble(prhs[i]) || mxIsComplex(prhs[i])) {
			string msg = string("Input argument #") + to_stringx(i + 1) + string(" is not double.");
			mexErrMsgTxt(msg.c_str());
			return;
        }
	}
    
    const double scale_factor = (nrhs >= 4) ? mxGetScalar(prhs[3]) : -1;
    const bool debug_mode = (nrhs >= 5) ? mxGetScalar(prhs[4]) : 0;

	const mxArray *sup = prhs[0];
	mwSize nsup = mxGetNumberOfElements(sup);
	const mxArray *dem = prhs[1];
	mwSize ndem = mxGetNumberOfElements(dem);
	const mxArray *D = prhs[2];

	if(mxGetNumberOfElements(D) != nsup * ndem) {
		mexErrMsgTxt("Ground distance matrix' size does not match the sizes of supply/demand vectors.");
		return;
	}

    if(debug_mode)
        mexPrintf("DBG: nsup = %d, ndem = %d\n", nsup, ndem);

    string input_fname(tmpnam(NULL));
	ofstream input_s(input_fname.c_str());
    if(debug_mode)
        mexPrintf("DBG: input_fname = '%s'\n", input_fname.c_str());
	trans_to_dimacs(input_s, sup, dem, D);
    input_s.close();
    if(debug_mode)
        mexPrintf("DBG: DIMACS-formatted input written\n");
    
	string raw_cs2_output = run_cs2(input_fname, scale_factor, nlhs > 1, debug_mode);
    if(!debug_mode)
        remove(input_fname.c_str());
    else
        mexPrintf("DBG: Raw CS2 output follows\n\n-----\n%s\n-----\n", raw_cs2_output.c_str());
    
    bool success = parse_cs2_output(raw_cs2_output, &plhs[0], &plhs[1], nlhs);
    if(!success && debug_mode)
        mexPrintf("Unable to parse CS2's output. Input file path: %s\n", input_fname.c_str());
}

bool parse_cs2_output(const string &cs2_raw_output, mxArray** plhs0, mxArray** plhs1, uint64_t nlhs) {
    bool success = true;
    double cost = -1;
    do {
        int64_t lpos = cs2_raw_output.find("cost:");
        if(lpos < 0) {
            mexWarnMsgTxt("Unable to find cost in the raw output of CS2.");
            mexWarnMsgTxt(((string)"Raw output is written to '" + write_str_to_tmpfile(cs2_raw_output) + "'").c_str());
            success = false;
            break;
        }
        lpos += strlen("cost:");
        int64_t rpos = cs2_raw_output.find("\n", lpos + 1);
        if(rpos < 0) {
            mexWarnMsgTxt("Raw output of CS2 seems to be malformed.");
            mexPrintf("\n%s\n", cs2_raw_output.c_str());
            success = false;
            break;
        }
        string ss(cs2_raw_output.substr(lpos, rpos - lpos));
        cost = stodx(ss);
    } while(0);
    *plhs0 = mxCreateDoubleScalar(cost);
    
    if(nlhs > 1) {
        do {

            int64_t lpos = cs2_raw_output.find("c --flow-->\n") + ((string)"c --flow-->\n").length();
            if(lpos < 0) {
                mexWarnMsgTxt("Unable to find flow in raw output of CS2.");
                mexPrintf("\n%s\n", cs2_raw_output.c_str());
                success = false;
                break;
            }

            int64_t rpos = cs2_raw_output.find("c <--flow--\n") - 1;
            if(rpos < 0 || rpos < lpos) {
                mexWarnMsgTxt("Unable to find flow in raw output of CS2.");
                mexPrintf("\n%s\n", cs2_raw_output.c_str());
                success = false;
                break;
            }
            
            vector<double> irows, icols, vals;
            
            stringstream ss(cs2_raw_output.substr(lpos, rpos - lpos + 1));
            string line;
            char ch;
            int64_t irow, icol;
            int64_t val;
            while(getline(ss, line)) {
                stringstream liness(line);
                liness >> ch >> irow >> icol >> val;
                irows.push_back(irow);
                icols.push_back(icol);
                vals.push_back(val);
            }
            int64_t nrows = irows.size();

            *plhs1 = mxCreateDoubleMatrix(nrows, 3, mxREAL);
            double *p = mxGetPr(*plhs1);
            for(int64_t i = 0; i < nrows; ++i) {
                p[i] = irows[i];
                p[nrows + i] = icols[i];
                p[(nrows << 1) + i] = vals[i];
            }

        } while(0);
    }
    
    return success;
}

string run_cs2(const string &input_fname, const double scale_factor, bool need_flow, bool debug_mode) {
	string mex_dirpath = get_mex_dirpath();
	string cmd = mex_dirpath + "/cs2-4.3/cs2.exe ";
    if(scale_factor > 0) { cmd += to_stringx(scale_factor); } else { cmd += " 12 "; }
    if(need_flow) { cmd += " flow "; } else { cmd += " no-flow "; }
    if(debug_mode) { cmd += " debug "; } else { cmd += " no-debug "; }
    cmd += " < " + input_fname + " 2>&1";
    
    if(debug_mode)
        mexPrintf("Running CS2 via '%s'\n", cmd.c_str());
    
	FILE *pipe = popen(cmd.c_str(), "r");
	if(!pipe) {
		mexErrMsgTxt("Unable to create a pipe to run CS2.");
		return "";
	}

	stringstream ss;
	char buffer[256];
	while(fgets(buffer, 255, pipe)) { ss << buffer; }

	pclose(pipe);

	return ss.str();
}

void trans_to_dimacs(ofstream &os, const mxArray *sup, const mxArray *dem, const mxArray *D) {
	mwSize nsup = mxGetNumberOfElements(sup);
	mwSize ndem = mxGetNumberOfElements(dem);

    os << "p min " << nsup + ndem << " " << nsup * ndem << "\n";

    double maxcap = 0;
    uint64_t inode = 1;
    for(double *ptr = mxGetPr(sup); inode <= nsup; ) {
        uint64_t val = d2uint64(*ptr);
        os << "n " << inode << "  " << val << "\n";
        maxcap += val;
        ++inode;
        ++ptr;
    }
    uint64_t imaxcap = d2uint64(maxcap);

    for(double *ptr = mxGetPr(dem); inode <= ndem + nsup; ) {
        uint64_t val = d2uint64(*ptr);
        os << "n " << inode << " -" << val << "\n";
        ++inode;
        ++ptr;
    }
    
    double *ptr = mxGetPr(D);
    for(uint64_t i = 0; i < nsup; ++i) {
        for(uint64_t j = 0; j < ndem; ++j) {
            uint64_t cost = d2uint64(*(ptr + j * nsup + i));
            os << "a " << (i + 1) << " " << (nsup + j + 1) << " 0 " << imaxcap << " " << cost << "\n";
        }
    }
}

string get_mex_dirpath() {
	mxArray *lhs[1];
	mxArray *rhs[1];

	rhs[0] = mxCreateString("fullpath");	
	mexCallMATLAB(1, lhs, 1, rhs, "mfilename");

	char *c_fpath = mxArrayToString(lhs[0]);
	string fpath(c_fpath);
	mxFree(c_fpath);

	mxDestroyArray(lhs[0]);
	mxDestroyArray(rhs[0]);

	return fpath.substr(0, fpath.find_last_of("\\/"));
}

string write_str_to_tmpfile(string text) {
    string fpath(tmpnam(NULL));
	ofstream os(fpath.c_str());
    os << text;
	os.close();
    return fpath;
}

template<typename T>
inline std::string to_stringx(T x) {
    ostringstream ss;
    ss << x;
    return ss.str();
}

inline double stodx(string str) {
    return strtod(str.c_str(), NULL);
}

inline uint64_t d2uint64(double x) {
    return round(abs(x));
}
